package sys_test

import (
	"math/rand/v2"
	"testing"

	"git.ufz.de/ecoepi/asfmm/comps"
	"git.ufz.de/ecoepi/asfmm/params"
	"git.ufz.de/ecoepi/asfmm/sys"
	"github.com/mlange-42/ark-tools/app"
	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark-tools/system"
	"github.com/mlange-42/ark/ecs"
	"github.com/stretchr/testify/assert"
)

func TestInitHerds(t *testing.T) {

	model := app.New() //.Seed(42)
	p := params.DefaultParams()
	ecs.AddResource(&model.World, &p)

	model.AddSystem(&sys.InitHerds{})
	model.AddSystem(&system.FixedTermination{Steps: 1})

	model.Run()

	// TODO: ID necessary at all in model?
	farmFilter := ecs.NewFilter1[comps.HerdID](&model.World)
	query := farmFilter.Query()

	assert.GreaterOrEqual(t, query.Count(), 40000, "The number of farms should be 40000 or 40001 with default parameters") // int(float64(p.AreaSize*p.AreaSize) * p.FarmDensity)
	assert.LessOrEqual(t, query.Count(), 40001, "The number of farms should be 40000 or 40001 with default parameters")    // int(float64(p.AreaSize*p.AreaSize) * p.FarmDensity)

	seenIDs := make(map[int]bool)

	for query.Next() {
		id := query.Get()

		_, keyExists := seenIDs[id.Value]
		assert.Falsef(t, keyExists, "Duplicate ID found: %d", id.Value)

		seenIDs[id.Value] = true
	}

	// test for entities with equal Position if they are of different types
	positionFilter := ecs.NewFilter2[comps.HerdPosition, comps.HerdType](&model.World)
	positions := make(map[comps.HerdPosition]comps.HerdType)

	positionQuery := positionFilter.Query()
	for positionQuery.Next() {
		position, herdType := positionQuery.Get()

		if existingHerdType, exists := positions[*position]; exists {
			assert.NotEqualf(t, existingHerdType, herdType, "Entities with the same Position must not have the same HerdType: %v", position)
		} else {
			positions[*position] = *herdType
		}
	}
}

func TestCalculateNumPigs(t *testing.T) {

	model := app.New() //.Seed(42)

	randSrc := ecs.GetResource[resource.Rand](&model.World)
	rng := rand.New(randSrc)

	// Test the deterministic case
	assert.Equal(t, 1000, sys.CalculateNumPigs(1000, 10, rng, 0.0), "With 0 variability, should return exactly the given mean value")

	// Test the random case with multiple iterations
	for range 1000 {
		numPigs := sys.CalculateNumPigs(1000, 10, rng, 0.5)
		assert.GreaterOrEqual(t, numPigs, 500, "Number of pigs should be at least 500")
		assert.LessOrEqual(t, numPigs, 1500, "Number of pigs should be at most 1500")
	}
}
