package sys

import (
	"math"

	"git.ufz.de/ecoepi/asfmm/comps"
	"git.ufz.de/ecoepi/asfmm/globals"
	"git.ufz.de/ecoepi/asfmm/params"
	"github.com/mlange-42/ark-tools/resource"
	"github.com/mlange-42/ark/ecs"
)

// ReleaseVirusCenter for virus release at the center of the simulation world
type ReleaseVirusCenter struct {
	tickRes                 ecs.Resource[resource.Tick]
	params                  ecs.Resource[params.Parameters]
	globalsWorld            ecs.Resource[globals.WorldGlobals]
	farmCenterDiseaseStatus *ecs.Map1[comps.HerdDiseaseStatus]
	centerHerds             []ecs.Entity
}

// Initialize find and saves herds, which are closest to the center of the simulation world.
func (s *ReleaseVirusCenter) Initialize(w *ecs.World) {
	s.tickRes = ecs.NewResource[resource.Tick](w)
	s.params = ecs.NewResource[params.Parameters](w)
	s.globalsWorld = ecs.NewResource[globals.WorldGlobals](w)
	s.farmCenterDiseaseStatus = ecs.NewMap1[comps.HerdDiseaseStatus](w)

	farmFilter := ecs.NewFilter1[comps.HerdPosition](w)
	linkedHerd := ecs.NewMap1[comps.LinkedHerd](w)

	centerX, centerY := float64(s.params.Get().World.SquareSideLength/2), float64(s.params.Get().World.SquareSideLength/2)
	lowestDistance := math.MaxFloat64
	s.centerHerds = []ecs.Entity{}

	herdCenter := ecs.Entity{}

	query := farmFilter.Query()
	for query.Next() {
		position := query.Get()
		distance := math.Sqrt(math.Pow(position.X-centerX, 2) + math.Pow(position.Y-centerY, 2))

		if distance < lowestDistance {
			herdCenter = query.Entity()
			lowestDistance = distance
		}
	}
	s.centerHerds = append(s.centerHerds, herdCenter)

	// do again for linked herd
	linked := linkedHerd.Get(herdCenter)
	if linked != nil {
		s.centerHerds = append(s.centerHerds, linked.Herd)
	}
}

// Update releases virus in herds at the center of the simulation world
func (s *ReleaseVirusCenter) Update(w *ecs.World) {
	tick := uint64(s.tickRes.Get().Tick)

	if tick != s.params.Get().Disease.VirusReleaseDay {
		return
	}

	for _, farmCenter := range s.centerHerds {
		status := s.farmCenterDiseaseStatus.Get(farmCenter)
		status.Status = comps.Exposed
		status.TickOfExposure = tick
	}

	// overwrite VirusReleased in world parameters
	// TODO: currently redundant global
	s.globalsWorld.Get().VirusReleased = true
}

// Finalize virus release
func (s *ReleaseVirusCenter) Finalize(w *ecs.World) {

}
